﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  /// <summary>
  /// This class implements the bottom-up tree distance described in the following paper: 
  /// G. Valiente, "An Efficient Bottom-up Distance Between Trees", http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.102.7958
  /// </summary>
  [StorableClass]
  [Item("Bottom-up distance calculator", "An operator that calculates the bottom-up distance between two symbolic expression trees.")]
  public class BottomUpTreeDistanceCalculator : Item, ISymbolicExpressionTreeDistanceCalculator {
    public BottomUpTreeDistanceCalculator() { }

    protected BottomUpTreeDistanceCalculator(BottomUpTreeDistanceCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BottomUpTreeDistanceCalculator(this, cloner);
    }

    // return a normalized distance measure in the interval (0,1)
    public double CalculateDistance(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      return BottomUpDistance(t1, t2) / (t1.Length + t2.Length);
    }

    public static double BottomUpDistance(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      var compactedGraph = Compact(t1, t2);

      var forwardMap = new Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode>(); // nodes of t1 => nodes of t2
      var reverseMap = new Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode>(); // nodes of t2 => nodes of t1

      foreach (var v in t1.IterateNodesBreadth()) {
        if (forwardMap.ContainsKey(v)) continue;
        var kv = compactedGraph[v];
        var w = t2.IterateNodesBreadth().FirstOrDefault(x => !reverseMap.ContainsKey(x) && compactedGraph[x] == kv); // not sure of iteration order here (pre/post order or breadth), this seems to work
        if (w == null) continue;

        // at this point we know that v and w are isomorphic, however, the mapping cannot be done directly (as in the paper) because the trees are unordered (subtree order might differ)
        // the solution is to sort subtrees by label using IterateBreadthOrdered (this will work because the subtrees are isomorphic!) and simultaneously iterate over the two subtrees
        var eV = IterateBreadthOrdered(v).GetEnumerator();
        var eW = IterateBreadthOrdered(w).GetEnumerator();

        while (eV.MoveNext() && eW.MoveNext()) {
          var s = eV.Current;
          var t = eW.Current;
          forwardMap[s] = t;
          reverseMap[t] = s;
        }
      }

      int c = forwardMap.Count(x => !Label(x.Key).Equals(Label(x.Value)));
      double distance = c + t1.Length + t2.Length - 2 * forwardMap.Count;
      return distance;
    }

    /// <summary>
    /// Creates a compact representation of the two trees as a directed acyclic graph
    /// </summary>
    /// <param name="t1">The first tree</param>
    /// <param name="t2">The second tree</param>
    /// <returns>The compacted DAG representing the two trees</returns>
    private static Dictionary<ISymbolicExpressionTreeNode, IVertex> Compact(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      var nodesToVertices = new Dictionary<ISymbolicExpressionTreeNode, IVertex>(); // K
      var disjointUnion = new DisjointUnion(t1, t2); // F
      var labelsToVertices = new Dictionary<string, IVertex>(); // L
      var childrenCount = new Dictionary<ISymbolicExpressionTreeNode, int>(); // Children
      var vertices = new List<IVertex>(); // G

      var nodes = disjointUnion.Nodes.ToList();

      // for all leaf labels l in F
      foreach (var l in nodes.Where(x => x.SubtreeCount == 0)) {
        var label = Label(l);
        var z = new Vertex { Content = l, Label = label };
        labelsToVertices[z.Label] = z;
        vertices.Add(z);
      }

      var treeNodes = new Queue<ISymbolicExpressionTreeNode>();

      foreach (var v in nodes) {
        childrenCount[v] = v.SubtreeCount;
        if (v.SubtreeCount == 0) {
          treeNodes.Enqueue(v);
        }
      }

      while (treeNodes.Any()) {
        var v = treeNodes.Dequeue();
        var label = Label(v);
        if (v.SubtreeCount == 0) {
          nodesToVertices[v] = labelsToVertices[label]; // 18
        } else {
          bool found = false;
          // for all nodes w in G in reverse order
          for (int i = vertices.Count - 1; i >= 0; --i) {
            var w = vertices[i];
            if (Height(v) != Height(w) || v.SubtreeCount != w.OutDegree || label != w.Label)
              continue;

            // sort V and W because we are dealing with unordered trees
            var vSubtrees = v.Subtrees.Select(s => nodesToVertices[s]).OrderBy(x => x.Label);
            var wSubtrees = w.OutArcs.Select(x => x.Target).OrderBy(x => x.Label);
            if (vSubtrees.SequenceEqual(wSubtrees)) {
              nodesToVertices[v] = w;
              found = true;
              break;
            }
          } // 32: end for

          if (!found) {
            var w = new Vertex { Content = v, Label = label };
            vertices.Add(w);
            nodesToVertices[v] = w;

            foreach (var u in v.Subtrees) {
              AddArc(w, nodesToVertices[u]);
            } // 40: end for
          } // 41: end if
        } // 42: end if

        if (v.Parent != null) {
          var p = v.Parent;
          childrenCount[p]--;
          if (childrenCount[p] == 0)
            treeNodes.Enqueue(p);
        }
      };

      return nodesToVertices;
    }

    private static IEnumerable<ISymbolicExpressionTreeNode> IterateBreadthOrdered(ISymbolicExpressionTreeNode node) {
      var list = new List<ISymbolicExpressionTreeNode> { node };
      int i = 0;
      while (i < list.Count) {
        var n = list[i];
        if (n.SubtreeCount > 0)
          list.AddRange(n.Subtrees.OrderBy(Label));
        i++;
      }
      return list;
    }

    private static IArc AddArc(IVertex source, IVertex target) {
      var arc = new Arc(source, target);
      source.AddForwardArc(arc);
      target.AddReverseArc(arc);
      return arc;
    }


    private static string Label(ISymbolicExpressionTreeNode n) {
      return n.ToString();
    }

    private static int Height(IVertex v) {
      return ((ISymbolicExpressionTreeNode)v.Content).GetDepth();
    }

    private static int Height(ISymbolicExpressionTreeNode n) {
      return n.GetDepth();
    }

    private class DisjointUnion : Tuple<ISymbolicExpressionTree, ISymbolicExpressionTree> {
      public DisjointUnion(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2)
        : base(t1, t2) {
      }

      public IEnumerable<ISymbolicExpressionTreeNode> Nodes {
        get { return Item1.Root.IterateNodesPostfix().Concat(Item2.Root.IterateNodesPostfix()); }
      }
    }

    public static string RemoveFirstLines(string text, int linesCount) {
      var lines = Regex.Split(text, "\r\n|\r|\n").Skip(linesCount);
      return string.Join(Environment.NewLine, lines.ToArray());
    }

    // draw the mapping between t1 and t2 as a tikz picture
    private static string FormatMapping(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2, Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode> map) {
      var symbolNameMap = new Dictionary<string, string>
    {
      {"ProgramRootSymbol", "Prog"},
      {"StartSymbol","RPB"},
      {"Multiplication", "$\\times$"},
      {"Division", "$\\div$"},
      {"Addition", "$+$"},
      {"Subtraction", "$-$"},
      {"Exponential", "$\\exp$"},
      {"Logarithm", "$\\log$"}
    };

      var sb = new StringBuilder();
      var nodeIds = new Dictionary<ISymbolicExpressionTreeNode, string>();
      int offset = 0;
      var layoutEngine = new ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode>(x => x.Subtrees);
      var nodeCoordinates = layoutEngine.CalculateLayout(t1.Root).ToDictionary(n => n.Content, n => new PointF(n.X, n.Y));

      double ws = 0.5;
      double hs = 0.5;

      var nl = Environment.NewLine;
      sb.Append("\\documentclass[class=minimal,border=0pt]{standalone}" + nl +
                 "\\usepackage{tikz}" + nl +
                 "\\begin{document}" + nl +
                 "\\begin{tikzpicture}" + nl +
                 "\\def\\ws{1}" + nl +
                 "\\def\\hs{0.7}" + nl +
                 "\\def\\offs{" + offset + "}" + nl);

      foreach (var node in t1.IterateNodesBreadth()) {
        var id = Guid.NewGuid().ToString();
        nodeIds[node] = id;
        var coord = nodeCoordinates[node];
        var nodeName = symbolNameMap.ContainsKey(node.Symbol.Name) ? symbolNameMap[node.Symbol.Name] : node.ToString();
        sb.AppendLine(string.Format(CultureInfo.InvariantCulture, "\\node ({0}) at (\\ws*{1} + \\offs,\\hs*{2}) {{{3}}};", nodeIds[node], ws * coord.X, -hs * coord.Y, EscapeLatexString(nodeName)));
      }

      foreach (ISymbolicExpressionTreeNode t in t1.IterateNodesBreadth()) {
        var n = t;
        foreach (var s in t.Subtrees) {
          sb.AppendLine(string.Format(CultureInfo.InvariantCulture, "\\draw ({0}) -- ({1});", nodeIds[n], nodeIds[s]));
        }
      }

      nodeCoordinates = layoutEngine.CalculateLayout(t2.Root).ToDictionary(n => n.Content, n => new PointF(n.X, n.Y));

      offset = 20;
      sb.Append("\\def\\offs{" + offset + "}" + nl);
      foreach (var node in t2.IterateNodesBreadth()) {
        var id = Guid.NewGuid().ToString();
        nodeIds[node] = id;
        var coord = nodeCoordinates[node];
        var nodeName = symbolNameMap.ContainsKey(node.Symbol.Name) ? symbolNameMap[node.Symbol.Name] : node.ToString();
        sb.AppendLine(string.Format(CultureInfo.InvariantCulture, "\\node ({0}) at (\\ws*{1} + \\offs,\\hs*{2}) {{{3}}};", nodeIds[node], ws * coord.X, -hs * coord.Y, EscapeLatexString(nodeName)));
      }

      foreach (ISymbolicExpressionTreeNode t in t2.IterateNodesBreadth()) {
        var n = t;
        foreach (var s in t.Subtrees) {
          sb.AppendLine(string.Format(CultureInfo.InvariantCulture, "\\draw ({0}) -- ({1});", nodeIds[n], nodeIds[s]));
        }
      }

      foreach (var p in map) {
        var id1 = nodeIds[p.Key];
        var id2 = nodeIds[p.Value];

        sb.Append(string.Format(CultureInfo.InvariantCulture, "\\path[draw,->,color=gray] ({0}) edge[bend left,dashed] ({1});" + Environment.NewLine, id1, id2));
      }
      sb.Append("\\end{tikzpicture}" + nl +
                "\\end{document}" + nl);
      return sb.ToString();
    }

    private static string EscapeLatexString(string s) {
      return s.Replace("\\", "\\\\").Replace("{", "\\{").Replace("}", "\\}").Replace("_", "\\_");
    }
  }
}
