﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  [StorableClass]
  [Item("Isomorphic tree distance calculator", "An operator that computes the distance between two symbolic expression trees based on the size of their largest common subtree.")]
  public class IsomorphicTreeDistanceCalculator : Item, ISymbolicExpressionTreeDistanceCalculator {
    private readonly SymbolicExpressionTreeNodeSimilarityComparer comparer;

    public IsomorphicTreeDistanceCalculator() {
      comparer = new SymbolicExpressionTreeNodeSimilarityComparer {
        MatchConstantValues = true,
        MatchVariableNames = true,
        MatchVariableWeights = true
      };
    }

    protected IsomorphicTreeDistanceCalculator(IsomorphicTreeDistanceCalculator original, Cloner cloner)
      : base(original, cloner) {
      this.comparer = original.comparer;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new IsomorphicTreeDistanceCalculator(this, cloner);
    }

    public double CalculateDistance(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      return 1 - MaxCommonSubtreeSimilarity(t1, t2, comparer); // we return the complement of the similarity to resemble the notion of a "distance"
    }

    public static double MaxCommonSubtreeSimilarity(ISymbolicExpressionTree a, ISymbolicExpressionTree b, SymbolicExpressionTreeNodeSimilarityComparer comparer) {
      int max = 0;
      var rootA = a.Root.GetSubtree(0).GetSubtree(0);
      var rootB = b.Root.GetSubtree(0).GetSubtree(0);
      foreach (var aa in rootA.IterateNodesBreadth()) {
        int lenA = aa.GetLength();
        if (lenA <= max) continue;
        foreach (var bb in rootB.IterateNodesBreadth()) {
          int lenB = bb.GetLength();
          if (lenB <= max) continue;
          int matches = SymbolicExpressionTreeMatching.Match(aa, bb, comparer);
          if (max < matches) max = matches;
        }
      }
      return 2.0 * max / (rootA.GetLength() + rootB.GetLength());
    }
  }
}
